/*
 * Amazon FreeRTOS V1.4.7
 * Copyright (C) 2017 Amazon.com, Inc. or its affiliates.  All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * http://aws.amazon.com/freertos
 * http://www.FreeRTOS.org
 */

/* FIX ME: If you are using Ethernet network connections and the FreeRTOS+TCP stack,
 * uncomment the define:
 * #define CY_USE_FREERTOS_TCP 1
 */

/* FreeRTOS includes. */
#include "FreeRTOS.h"
#include "task.h"

#ifdef CY_USE_LWIP
#include "tcpip.h"
#endif

/* Demo includes */
#include "aws_demo.h"

/* AWS library includes. */
#include "iot_system_init.h"
#include "iot_logging_task.h"
#include "iot_wifi.h"
#include "aws_clientcredential.h"
#include "aws_application_version.h"
#include "aws_dev_mode_key_provisioning.h"
/* Logging includes. */
#include "private/iot_logging.h"

/* BSP & Abstraction inclues */
#include "cybsp.h"
#include "cy_retarget_io.h"

#include "cy_pdl.h"
#include "cycfg.h"
#include "cyabs_rtos.h"

#include "platform.h"
#include "platform_peripheral.h"
//#include "uartTask.h"
#include "cyhal_gpio.h"
#include "FreeRTOSConfig.h"

/* Logging Task Defines. */
//#define mainLOGGING_MESSAGE_QUEUE_LENGTH    ( 15 )
#define mainLOGGING_MESSAGE_QUEUE_LENGTH    ( 32 )
#define mainLOGGING_TASK_STACK_SIZE         ( configMINIMAL_STACK_SIZE * 8 )

/* The task delay for allowing the lower priority logging task to print out Wi-Fi
 * failure status before blocking indefinitely. */
#define mainLOGGING_WIFI_STATUS_DELAY       pdMS_TO_TICKS( 1000 )

/* Unit test defines. */
#define mainTEST_RUNNER_TASK_STACK_SIZE     ( configMINIMAL_STACK_SIZE * 16 )

/* The name of the devices for xApplicationDNSQueryHook. */
#define mainDEVICE_NICK_NAME				"cypress_demo" /* FIX ME.*/

static SemaphoreHandle_t uartSemaphore = NULL;
uint32_t read_data;
uint32_t read_element;
cy_stc_scb_uart_context_t CYBSP_AT_UART_context;

/* Static arrays for FreeRTOS-Plus-TCP stack initialization for Ethernet network
 * connections are declared below. If you are using an Ethernet connection on your MCU
 * device it is recommended to use the FreeRTOS+TCP stack. The default values are
 * defined in FreeRTOSConfig.h. */

#ifdef CY_USE_FREERTOS_TCP
/* Default MAC address configuration.  The application creates a virtual network
 * connection that uses this MAC address by accessing the raw Ethernet data
 * to and from a real network connection on the host PC.  See the
 * configNETWORK_INTERFACE_TO_USE definition for information on how to configure
 * the real network connection to use. */
const uint8_t ucMACAddress[ 6 ] =
{
    configMAC_ADDR0,
    configMAC_ADDR1,
    configMAC_ADDR2,
    configMAC_ADDR3,
    configMAC_ADDR4,
    configMAC_ADDR5
};

/* The default IP and MAC address used by the application.  The address configuration
 * defined here will be used if ipconfigUSE_DHCP is 0, or if ipconfigUSE_DHCP is
 * 1 but a DHCP server could not be contacted.  See the online documentation for
 * more information. */
static const uint8_t ucIPAddress[ 4 ] =
{
    configIP_ADDR0,
    configIP_ADDR1,
    configIP_ADDR2,
    configIP_ADDR3
};
static const uint8_t ucNetMask[ 4 ] =
{
    configNET_MASK0,
    configNET_MASK1,
    configNET_MASK2,
    configNET_MASK3
};
static const uint8_t ucGatewayAddress[ 4 ] =
{
    configGATEWAY_ADDR0,
    configGATEWAY_ADDR1,
    configGATEWAY_ADDR2,
    configGATEWAY_ADDR3
};
static const uint8_t ucDNSServerAddress[ 4 ] =
{
    configDNS_SERVER_ADDR0,
    configDNS_SERVER_ADDR1,
    configDNS_SERVER_ADDR2,
    configDNS_SERVER_ADDR3
};
#endif /* CY_USE_FREERTOS_TCP */

/**
 * @brief Application task startup hook for applications using Wi-Fi. If you are not
 * using Wi-Fi, then start network dependent applications in the vApplicationIPNetorkEventHook
 * function. If you are not using Wi-Fi, this hook can be disabled by setting
 * configUSE_DAEMON_TASK_STARTUP_HOOK to 0.
 */
void vApplicationDaemonTaskStartupHook( void );

/**
 * @brief Connects to Wi-Fi.
 */
static void prvWifiConnect( void );

/**
 * @brief Initializes the board.
 */
static void prvMiscInitialization( void );

const IotLogConfig_t azlogHideAll =
    {
        .hideLibraryName = true,
        .hideLogLevel    = true,
        .hideTimestring  = true
    };

#define RX_BUFFER_SIZE 256

void handle_error(void)
{
	/* Disable all interrupts */
	__disable_irq();

	/* Switch on error LED */
	//Cy_GPIO_Write(KIT_LED2_PORT, KIT_LED2_NUM, LED_ON);
	printf("uart error!!!!\n");
	//while(1u) {}
}
//#if 0
/* Create the Interrupt Config Structure, this is automatically generated in Creator, but not in MT */
const cy_stc_sysint_t CYBSP_AT_UART_IRQ_cfg = {
		.intrSrc = CYBSP_AT_UART_IRQ,
		.intrPriority = 7u,
		//.intrPriority = 3,
};

static void UART_Isr(){

    Cy_SCB_SetRxInterruptMask(CYBSP_AT_UART_HW, 0);
    Cy_SCB_ClearRxInterrupt(CYBSP_AT_UART_HW, CY_SCB_RX_INTR_NOT_EMPTY);
    NVIC_ClearPendingIRQ(CYBSP_AT_UART_IRQ_cfg.intrSrc);

    BaseType_t xHigherPriorityTaskWoken;
    xHigherPriorityTaskWoken = pdFALSE;

    xSemaphoreGiveFromISR(uartSemaphore, &xHigherPriorityTaskWoken);

    if(xHigherPriorityTaskWoken != pdFALSE){
        portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
    }
}
//#endif
int at_drv_uart_init(uint32_t baud)
{

	cy_stc_scb_uart_context_t CYBSP_AT_UART_context;
	cy_en_scb_uart_status_t initstatus;

	Cy_SCB_UART_Init(CYBSP_AT_UART_HW, &CYBSP_AT_UART_config, &CYBSP_AT_UART_context);

	Cy_SCB_UART_Enable(CYBSP_AT_UART_HW);
	setvbuf(stdin, NULL, _IONBF, 0);

	uartSemaphore = xSemaphoreCreateBinary();

	Cy_SysInt_Init(&CYBSP_AT_UART_IRQ_cfg, &UART_Isr);
	NVIC_EnableIRQ(CYBSP_AT_UART_IRQ_cfg.intrSrc);
	Cy_SCB_SetRxInterruptMask(CYBSP_AT_UART_HW, CY_SCB_RX_INTR_NOT_EMPTY);
	//Cy_SCB_SetTxInterruptMask(CYBSP_AT_UART_HW, CY_SCB_TX_INTR_LEVEL);

    //Cy_SCB_UART_PutString(CYBSP_AT_UART_HW, "at_uart_init.\r\n");

	return CY_SCB_UART_SUCCESS;
}

int at_drv_uart_deinit()
{

	Cy_SCB_UART_Disable(CYBSP_AT_UART_HW, &CYBSP_AT_UART_context);
	Cy_SCB_UART_DeInit(CYBSP_AT_UART_HW);
	return CY_SCB_UART_SUCCESS;
}

int at_uart_read(uint8_t *data_buf, int len)
{
	int l_len = 0;

	while(l_len < len) {
		//configPRINTF( ( "prepare to read.\r\n" ) );
		xSemaphoreTake(uartSemaphore, portMAX_DELAY);

	    l_len += Cy_SCB_UART_GetArray(CYBSP_AT_UART_HW, data_buf + l_len, len - l_len);

	    Cy_SCB_SetRxInterruptMask(CYBSP_AT_UART_HW, CY_SCB_RX_INTR_NOT_EMPTY);

	}

	//configPRINTF( ( "handling command:[%c]\r\n", data_buf[0] ) );

	return l_len;
}

int at_uart_write(uint8_t *data_buf, int len)
{
	int l_len = 0;
	while (l_len < len) {
		l_len += Cy_SCB_UART_PutArray(CYBSP_AT_UART_HW,
			data_buf + l_len, len - l_len);

	}

	return l_len;
}

int at_uart_init()
{
	return at_drv_uart_init();
}

int at_uart_deinit()
{
	return at_drv_uart_deinit();
}

int at_uart_putchar(uint8_t *data)
{
	//configPRINTF( ( "=%c\r\n", *data ) );
	int len = at_uart_write(data, 1);
	return len;
}

int at_uart_getchar(uint8_t *data)
{
	int len = at_uart_read(data, 1);
	return len;
}

int psm_read(const uint8_t *name, void *buf, uint32_t size)
{
	return 0;
}

int psm_write(const uint8_t *name, const void *buf, uint32_t size)
{
	return 0;
}

int at_gpio_read(GPIO_PRT_Type* base, uint32_t pinNum)
{
	return Cy_GPIO_Read(base, pinNum);
}

void at_gpio_write(GPIO_PRT_Type* base, uint32_t pinNum, uint32_t value)
{
	Cy_GPIO_Write(base, pinNum, value);
}

void at_reset(void)
{
	configPRINTF( ( "prepare to reset.\r\n" ) );
	NVIC_SystemReset();
}
/*-----------------------------------------------------------*/

/**
 * @brief Application runtime entry point.
 */
int main( void )
{
    /* Perform any hardware initialization that does not require the RTOS to be
     * running.  */
    prvMiscInitialization();

    /* Create tasks that are not dependent on the Wi-Fi being initialized. */
    xLoggingTaskInitialize( mainLOGGING_TASK_STACK_SIZE,
                            tskIDLE_PRIORITY,
                            mainLOGGING_MESSAGE_QUEUE_LENGTH );

#ifdef CY_USE_FREERTOS_TCP
    FreeRTOS_IPInit( ucIPAddress,
                     ucNetMask,
                     ucGatewayAddress,
                     ucDNSServerAddress,
                     ucMACAddress );
#endif /* CY_USE_FREERTOS_TCP */
    init_cycfg_all();

    __enable_irq();

    /* Start the scheduler.  Initialization that requires the OS to be running,
     * including the Wi-Fi initialization, is performed in the RTOS daemon task
     * startup hook. */
    vTaskStartScheduler();

    return 0;
}
/*-----------------------------------------------------------*/

/*-----------------------------------------------------------*/

static void prvMiscInitialization( void )
{
    cy_rslt_t result = cybsp_init();
    if (result != CY_RSLT_SUCCESS)
    {
        configPRINTF( (  "BSP initialization failed \r\n" ) );
    }
    result = cy_retarget_io_init(CYBSP_DEBUG_UART_TX, CYBSP_DEBUG_UART_RX, CY_RETARGET_IO_BAUDRATE);
    if (result != CY_RSLT_SUCCESS)
    {
        configPRINTF( ( "Retarget IO initializatoin failed \r\n" ) );
    }

}
/*-----------------------------------------------------------*/
void vApplicationDaemonTaskStartupHook( void )
{
    /* FIX ME: Perform any hardware initialization, that require the RTOS to be
     * running, here. */
	 //at_uart_init();
	 printf("Enter ApplicationDaemon!!!!\n");

    /* FIX ME: If your MCU is using Wi-Fi, delete surrounding compiler directives to
     * enable the unit tests and after MQTT, Bufferpool, and Secure Sockets libraries
     * have been imported into the project. If you are not using Wi-Fi, see the
     * vApplicationIPNetworkEventHook function. */
    if( SYSTEM_Init() == pdPASS )
    {
#ifdef CY_USE_LWIP
        /* Initialize lwIP stack. This needs the RTOS to be up since this function will spawn 
         * the tcp_ip thread.
         */
        tcpip_init(NULL, NULL);
#endif
        /* Connect to the Wi-Fi before running the tests. */
        //prvWifiConnect();

        cyhal_gpio_configure(INT_HAL_PORT_PIN, CYHAL_GPIO_DIR_OUTPUT, CYHAL_GPIO_DRIVE_STRONG);
        cyhal_gpio_configure(MSG_HAL_PORT_PIN, CYHAL_GPIO_DIR_INPUT, CYHAL_GPIO_DRIVE_PULLDOWN);

        Cy_GPIO_Write(INT_PORT, INT_PIN, 1);
        int status = at_gpio_read(MSG_PORT, MSG_PIN);
        printf("status = %d!!!!\n", status);

        //Cy_GPIO_Write(INT_PORT, INT_PIN, 0);
        //vTaskDelay( pdMS_TO_TICKS( 1000 ) );
        //Cy_GPIO_Write(INT_PORT, INT_PIN, 1);
        //vTaskDelay( pdMS_TO_TICKS( 1000 ) );
        //Cy_GPIO_Write(INT_PORT, INT_PIN, 0);

        /* Provision the device with AWS certificate and private key. */
        vDevModeKeyProvisioning();

		printf("Preapre to demo tasks");
        /* Start the demo tasks. */
        DEMO_RUNNER_RunDemos();
    }
}
/*-----------------------------------------------------------*/

void prvWifiConnect( void )
{
    WIFINetworkParams_t xNetworkParams;
    WIFIReturnCode_t xWifiStatus;
    uint8_t ucTempIp[4] = { 0 };

    xWifiStatus = WIFI_On();

    if( xWifiStatus == eWiFiSuccess )
    {

        configPRINTF( ( "Wi-Fi module initialized. Connecting to AP...\r\n" ) );
    }
    else
    {
        configPRINTF( ( "Wi-Fi module failed to initialize.\r\n" ) );

        /* Delay to allow the lower priority logging task to print the above status.
            * The while loop below will block the above printing. */
        vTaskDelay( mainLOGGING_WIFI_STATUS_DELAY );

        while( 1 )
        {
        }
    }

    /* Setup parameters. */
    xNetworkParams.pcSSID = clientcredentialWIFI_SSID;
    xNetworkParams.ucSSIDLength = sizeof( clientcredentialWIFI_SSID ) - 1;
    xNetworkParams.pcPassword = clientcredentialWIFI_PASSWORD;
    xNetworkParams.ucPasswordLength = sizeof( clientcredentialWIFI_PASSWORD ) - 1;
    xNetworkParams.xSecurity = clientcredentialWIFI_SECURITY;
    xNetworkParams.cChannel = 0;

    xWifiStatus = WIFI_ConnectAP( &( xNetworkParams ) );

    if( xWifiStatus == eWiFiSuccess )
    {
        configPRINTF( ( "Wi-Fi Connected to AP. Creating tasks which use network...\r\n" ) );

        xWifiStatus = WIFI_GetIP( ucTempIp );
        if ( eWiFiSuccess == xWifiStatus )
        {
            configPRINTF( ( "IP Address acquired %d.%d.%d.%d\r\n",
                            ucTempIp[ 0 ], ucTempIp[ 1 ], ucTempIp[ 2 ], ucTempIp[ 3 ] ) );
        }
    }
    else
    {
        /* Connection failed, configure SoftAP. */
        configPRINTF( ( "Wi-Fi failed to connect to AP %s.\r\n", xNetworkParams.pcSSID ) );

        xNetworkParams.pcSSID = wificonfigACCESS_POINT_SSID_PREFIX;
        xNetworkParams.pcPassword = wificonfigACCESS_POINT_PASSKEY;
        xNetworkParams.xSecurity = wificonfigACCESS_POINT_SECURITY;
        xNetworkParams.cChannel = wificonfigACCESS_POINT_CHANNEL;

        configPRINTF( ( "Connect to SoftAP %s using password %s. \r\n",
                        xNetworkParams.pcSSID, xNetworkParams.pcPassword ) );

        while( WIFI_ConfigureAP( &xNetworkParams ) != eWiFiSuccess )
        {
            configPRINTF( ( "Connect to SoftAP %s using password %s and configure Wi-Fi. \r\n",
                            xNetworkParams.pcSSID, xNetworkParams.pcPassword ) );
        }

        configPRINTF( ( "Wi-Fi configuration successful. \r\n" ) );
    }
}

/*-----------------------------------------------------------*/

/**
 * @brief User defined Idle task function.
 *
 * @note Do not make any blocking operations in this function.
 */
void vApplicationIdleHook( void )
{
    /* FIX ME. If necessary, update to application idle periodic actions. */

    static TickType_t xLastPrint = 0;
    TickType_t xTimeNow;
    const TickType_t xPrintFrequency = pdMS_TO_TICKS( 5000 );

    xTimeNow = xTaskGetTickCount();

    if( ( xTimeNow - xLastPrint ) > xPrintFrequency )
    {
        configPRINT( "." );
        xLastPrint = xTimeNow;
    }
}

void vApplicationTickHook()
{
}

/*-----------------------------------------------------------*/

/**
* @brief User defined application hook to process names returned by the DNS server.
*/
#if ( ipconfigUSE_LLMNR != 0 ) || ( ipconfigUSE_NBNS != 0 )
    BaseType_t xApplicationDNSQueryHook( const char * pcName )
    {
        /* FIX ME. If necessary, update to applicable DNS name lookup actions. */

        BaseType_t xReturn;

        /* Determine if a name lookup is for this node.  Two names are given
         * to this node: that returned by pcApplicationHostnameHook() and that set
         * by mainDEVICE_NICK_NAME. */
        if( strcmp( pcName, pcApplicationHostnameHook() ) == 0 )
        {
            xReturn = pdPASS;
        }
        else if( strcmp( pcName, mainDEVICE_NICK_NAME ) == 0 )
        {
            xReturn = pdPASS;
        }
        else
        {
            xReturn = pdFAIL;
        }

        return xReturn;
    }

#endif /* if ( ipconfigUSE_LLMNR != 0 ) || ( ipconfigUSE_NBNS != 0 ) */
/*-----------------------------------------------------------*/

/**
 * @brief User defined assertion call. This function is plugged into configASSERT.
 * See FreeRTOSConfig.h to define configASSERT to something different.
 */
void vAssertCalled(const char * pcFile,
	uint32_t ulLine)
{
    /* FIX ME. If necessary, update to applicable assertion routine actions. */

	const uint32_t ulLongSleep = 1000UL;
	volatile uint32_t ulBlockVariable = 0UL;
	volatile char * pcFileName = (volatile char *)pcFile;
	volatile uint32_t ulLineNumber = ulLine;

	(void)pcFileName;
	(void)ulLineNumber;

	printf("vAssertCalled %s, %ld\n", pcFile, (long)ulLine);
	fflush(stdout);

	/* Setting ulBlockVariable to a non-zero value in the debugger will allow
	* this function to be exited. */
	taskDISABLE_INTERRUPTS();
	{
		while (ulBlockVariable == 0UL)
		{
			vTaskDelay( pdMS_TO_TICKS( ulLongSleep ) );
		}
	}
	taskENABLE_INTERRUPTS();
}
/*-----------------------------------------------------------*/

/**
 * @brief User defined application hook need by the FreeRTOS-Plus-TCP library.
 */
#if ( ipconfigUSE_LLMNR != 0 ) || ( ipconfigUSE_NBNS != 0 ) || ( ipconfigDHCP_REGISTER_HOSTNAME == 1 )
    const char * pcApplicationHostnameHook(void)
    {
        /* FIX ME: If necessary, update to applicable registration name. */

        /* This function will be called during the DHCP: the machine will be registered
         * with an IP address plus this name. */
        return clientcredentialIOT_THING_NAME;
    }
	
#endif
